from typing import Tuple

import torch
import torch.nn as nn

from ...modules import ConvLayer, Up
from .merging import Merge

__all__ = ["Unet3pSkip"]


class Unet3pSkip(nn.ModuleDict):
    def __init__(
        self,
        stage_ix: int,
        dec_channels: Tuple[int, ...],
        skip_channels: Tuple[int, ...],
        dec_dims: Tuple[int, ...],
        hid_channels: int = 320,
        n_layers: int = 1,
        n_blocks: Tuple[int, ...] = (1,),
        short_skips: Tuple[str, ...] = ("residual",),
        block_types: Tuple[Tuple[str, ...], ...] = (("basic",),),
        activations: Tuple[Tuple[str, ...], ...] = (("relu",),),
        normalizations: Tuple[Tuple[str, ...], ...] = (("bn",),),
        convolutions: Tuple[Tuple[str, ...], ...] = (("conv",),),
        attentions: Tuple[Tuple[str, ...], ...] = ((None,),),
        kernel_sizes: Tuple[int, ...] = ((3,),),
        groups: Tuple[int, ...] = ((1,),),
        biases: Tuple[Tuple[bool, ...], ...] = ((False,),),
        preactivates: Tuple[Tuple[bool, ...], ...] = ((False,),),
        preattends: Tuple[Tuple[bool, ...], ...] = ((False,),),
        use_styles: Tuple[Tuple[bool, ...], ...] = ((False,),),
        expand_ratios: Tuple[float, float] = ((1.0,),),
        merge_policy: str = "sum",
        lite_version: bool = False,
        **kwargs,
    ) -> None:
        """U-net3+-like skip connection block with added flexibility.

        UNet 3+: A Full-Scale Connected UNet for Medical Image Segmentation
            - https://arxiv.org/abs/2004.08790

        Sets up a conv block for the upsampled feature map from the
        previous decoder stage and dynamically sets up the conv blocks
        for the outputs of encoder stages and previous decoder stages.
        The number of these conv blocks depend on the decoder stage ix.

        a lite version can be used where the decoder-to-decoder skips
        are skipped.

        Parameters
        ----------
            stage_ix : int
                The index number of the current decoder stage.
            dec_channels : Tuple[int, ...]
                The number of output channels in the decoder output stages. First elem
                is the number of channels in the encoder head or bottleneck.
            skip_channels : Tuple[int, ...]
                List of the number of channels in the encoder skip tensors.
            dec_dims : Tuple[int, ...]
                List of the heights/widths of each encoder/decoder feature map
                e.g. [256, 128, 64, 32, 16]. Assumption is that feature maps are square.
            hid_channels : int
                Number of output channels from this module.
            n_layers : int, default=1
                The number of conv layers inside one skip stage.
            n_blocks : Tuple[int, ...], default=(1, )
                Number of conv-blocks used at each layer of the skip connection.
            short_skips : str, default=("residual", )
                The short skip methods used inside the conv layers.
            block_types : Tuple[Tuple[str, ...], ...], default=(("basic",), )
                The type of the conv blocks in the conv blocks inside the layers.
            normalizations : Tuple[Tuple[str, ...], ...], default: (("bn",), )
                Normalization methods used in the conv blocks inside the conv layers.
            convolutions : Tuple[str, ...], default=(("conv",),)
                The convolution method. One of: "conv", "wsconv", "scaled_wsconv"
            activations : Tuple[Tuple[str, ...], ...], default: (("relu",), )
                Activation methods used inside the conv layers.
            kernel_sizes : Tuple[int, ...], default=(3, 3)
                The size of the convolution kernels in each conv block.
            groups : int, default=((1,),)
                Number of groups for the kernels in each convolution blocks.
            biases : bool, default=((False,),)
                Include bias terms in the convolution blocks.
            attentions : Tuple[Tuple[str, ...], ...], default: ((None,), )
                Attention methods used inside the conv layers.
            preactivates Tuple[Tuple[bool, ...], ...], default: ((False,), )
                Boolean flags for the conv layers to use pre-activation.
            preattends Tuple[Tuple[bool, ...], ...], default: ((False,), )
                Boolean flags for the conv layers to use pre-activation.
            use_styles : Tuple[Tuple[bool, ...], ...], default=((False,), )
                Boolean flags for the conv layers to add style vectors at each block.
            expand_ratios : Tuple[float, float], default=((1.0, ),)
                Expand ratios for the conv blocks.
            upsampling : str, default="fixed-unpool"
                Name of the upsampling method.
            merge_policy : str, default="sum"
                The long skip merge policy. One of: "sum", "cat"
            lite_version : bool, default=False
                If True, the dense decoder-to-decoder skips are not utilized at all.
                Reduces the model params quite a lot and computational cost.
        """
        super().__init__()

        illegal_args = [
            (k, a)
            for k, a in locals().items()
            if isinstance(a, tuple)
            and a not in (skip_channels, dec_channels, dec_dims)
            and len(a) != n_layers
        ]

        if illegal_args:
            raise ValueError(
                f"""
                All the tuple-arg lengths need to be equal to `n_layers`={n_layers}.
                Illegal args: {illegal_args}"""
            )

        if dec_channels[stage_ix] <= 4:
            raise ValueError(
                f"Input channels need to be larger than 4. Got {dec_channels[stage_ix]}"
            )

        self.stage_ix = stage_ix
        self.merge_policy = merge_policy
        self.lite_version = lite_version
        self.n_skips = len(skip_channels) + 1
        self.fin_channels = hid_channels
        self.encoder_skip_channels = skip_channels[stage_ix:]

        if stage_ix < len(skip_channels):
            decoder_out_dims = dec_dims[1:]
            decoder_skip_dims = dec_dims[:stage_ix]
            decoder_skip_channels = dec_channels[:stage_ix]
            encoder_skip_dims = decoder_out_dims[stage_ix:]
            target_size = decoder_out_dims[stage_ix]  # target size for pooling

            conv_channels = self.conv_channels
            out = conv_channels[0]  # out channels for the first conv op
            cat_channels = conv_channels[-1]  # out channels for the rest conv ops

            # conv-layer for the upsampled features
            for i in range(n_layers):
                layer = ConvLayer(
                    in_channels=dec_channels[stage_ix],
                    out_channels=out,
                    n_blocks=n_blocks[i],
                    expand_ratios=expand_ratios[i],
                    short_skip=short_skips[i],
                    block_types=block_types[i],
                    activations=activations[i],
                    normalizations=normalizations[i],
                    convolutions=convolutions[i],
                    attentions=attentions[i],
                    preattends=preattends[i],
                    preactivates=preactivates[i],
                    use_styles=use_styles[i],
                    kernel_sizes=kernel_sizes[i],
                    groups=groups[i],
                    biases=biases[i],
                )
                self.add_module("nonskip_layer", layer)

            # down-sampling blocks and conv-layers for the encoder2decoder connections
            if self.encoder_skip_channels:
                for i, (in_, d) in enumerate(
                    zip(self.encoder_skip_channels, encoder_skip_dims)
                ):
                    scale_op = self.get_scale_op(d, target_size)
                    self.add_module(f"enc2dec_downscale{i + 1}", scale_op)

                    for j in range(n_layers):
                        layer = ConvLayer(
                            in_channels=in_,
                            out_channels=cat_channels if merge_policy == "cat" else out,
                            n_blocks=n_blocks[j],
                            expand_ratios=expand_ratios[j],
                            short_skip=short_skips[j],
                            block_types=block_types[j],
                            activations=activations[j],
                            normalizations=normalizations[j],
                            convolutions=convolutions[j],
                            attentions=attentions[j],
                            preattends=preattends[j],
                            preactivates=preactivates[j],
                            use_styles=use_styles[j],
                            kernel_sizes=kernel_sizes[j],
                            groups=groups[j],
                            biases=biases[j],
                        )
                        self.add_module(f"enc2dec_layer{i + 1}", layer)

            # up-sampling blocks and conv-layers for the decoder2decoder connections
            if self.encoder_skip_channels and not lite_version:
                for i, (in_, d) in enumerate(
                    zip(decoder_skip_channels, decoder_skip_dims)
                ):

                    scale_op = self.get_scale_op(d, target_size)
                    self.add_module(f"dec2dec_upscale{i + 1}", scale_op)

                    for j in range(n_layers):
                        layer = ConvLayer(
                            in_channels=in_,
                            out_channels=cat_channels if merge_policy == "cat" else out,
                            n_blocks=n_blocks[j],
                            expand_ratios=expand_ratios[j],
                            short_skip=short_skips[j],
                            block_types=block_types[j],
                            activations=activations[j],
                            normalizations=normalizations[j],
                            convolutions=convolutions[j],
                            attentions=attentions[j],
                            preattends=preattends[j],
                            preactivates=preactivates[j],
                            use_styles=use_styles[j],
                            kernel_sizes=kernel_sizes[j],
                            groups=groups[j],
                            biases=biases[j],
                        )
                        self.add_module(f"dec2dec_layer{i + 1}", layer)

            # Final merge block
            merge = Merge(
                name=merge_policy,  # if stage_ix < self.n_skips - 1 else None,
                in_channels=self.nonskip_layer.out_channels,
                skip_channels=conv_channels,
            )
            self.add_module("merge_block", merge)
        else:
            # place-holder Merge-for the final decoder stage
            self.add_module(
                "merge_block",
                Merge(
                    None,
                    in_channels=dec_channels[stage_ix],
                    out_channels=dec_channels[stage_ix],
                ),
            )

    @property
    def out_channels(self) -> int:
        """Out channels."""
        if self.stage_ix < self.n_skips - 1:
            out_channels = self.fin_channels
        else:
            out_channels = self.merge_block.out_channels

        return out_channels

    @property
    def conv_channels(self) -> Tuple[int, ...]:
        """Convolution channel arithmetic.

        Returns
        -------
            Tuple[int, ...]:
                The number of channels for all the skip conv blocks.
        """
        # cat_channels = None
        out_channels = self.fin_channels
        divider = self.n_skips
        if self.lite_version:
            divider = len(self.encoder_skip_channels) + 1

        conv_channels = [self.fin_channels] * (divider - 1)

        if self.merge_policy == "cat":
            # divide the number of out channels evenly for each conv block
            cat_channels, reminder = divmod(self.fin_channels, divider)

            if self.encoder_skip_channels:
                out_channels = cat_channels + reminder

            conv_channels = [out_channels] + [cat_channels] * (divider - 1)

        return conv_channels

    @staticmethod
    def get_scale_op(in_size: int, target_size: int) -> nn.Module:
        """Get the up/down scaling operation for the feature maps."""
        scale_factor = in_size / target_size

        if scale_factor > 1:
            scale_op = nn.MaxPool2d(kernel_size=int(scale_factor), ceil_mode=True)
        elif scale_factor < 1:
            scale_op = Up("fixed-unpool", scale_factor=int(1 / scale_factor))
        else:
            scale_op = nn.Identity()

        return scale_op

    def forward(
        self,
        x: torch.Tensor,
        skips: Tuple[torch.Tensor],
        extra_skips: Tuple[torch.Tensor] = None,
        **kwargs,
    ) -> torch.Tensor:
        """Forward pass of the Unet3+ skip connection.

        Parameters
        ----------
            x : torch.Tensor
                Input tensor. Shape (B, C, H, W).
            skips : Tuple[torch.Tensor, ...]
                All of feature maps from consecutive encoder blocks.
                Order is bottom up. Shapes: (B, C, H, W).
            extra_skips : Tuple[torch.Tensor, ...], default=None
                Extra skip connections. (Decoder2decoder connections).

        Returns
        -------
            torch.Tensor:
                Output torch.Tensor. Shape: (B, C, H, W).
        """
        # init containers for feature maps
        encoder_features = []
        decoder_features = []

        if self.stage_ix < len(skips):
            skips = skips[self.stage_ix :]
            # forward pass for the non-skip layer
            x = self.nonskip_layer(x)

            # forward passes for the encoder2decoder connections
            enc2dec_scales = [k for k in self.keys() if "enc2dec_downscale" in k]
            enc2dec_layers = [k for k in self.keys() if "enc2dec_layer" in k]
            for i, (scale, layer) in enumerate(zip(enc2dec_scales, enc2dec_layers)):
                encoder_feat = self[scale](skips[i])
                encoder_feat = self[layer](encoder_feat)
                encoder_features.append(encoder_feat)

            # forward passes for the decoder2decoder connections
            dec2dec_scales = [k for k in self.keys() if "dec2dec_upscale" in k]
            dec2dec_layers = [k for k in self.keys() if "dec2dec_layer" in k]
            for i, (scale, layer) in enumerate(zip(dec2dec_scales, dec2dec_layers)):
                decoder_feat = self[scale](extra_skips[i])
                decoder_feat = self[layer](decoder_feat)
                decoder_features.append(decoder_feat)

        # Merge all the feature maps
        skip_features = encoder_features + decoder_features
        x = self.merge_block(x, skip_features)

        return x



import numpy as np
import pytest

from cellseg_models_pytorch.metrics import (
    accuracy_multiclass,
    dice_multiclass,
    f1score_multiclass,
    iou_multiclass,
    sensitivity_multiclass,
    specificity_multiclass,
)


@pytest.mark.parametrize("zero_true", [False, True])
@pytest.mark.parametrize("zero_pred", [False, True])
@pytest.mark.parametrize(
    "metric",
    [
        iou_multiclass,
        dice_multiclass,
        f1score_multiclass,
        accuracy_multiclass,
        sensitivity_multiclass,
        specificity_multiclass,
    ],
)
def test_semantic_metrics(true_sem, pred_sem, metric, zero_pred, zero_true):
    if zero_pred:
        pred_sem = np.zeros_like(true_sem)
    if zero_true:
        true_sem = np.zeros_like(pred_sem)

    metrics = metric(true_sem, pred_sem, num_classes=5)

    assert metrics.shape == (5,)

    if not zero_pred and not zero_true:
        assert metrics[2] != 1.0



import torch
import torch.nn as nn
from torch.nn.parameter import Parameter


# Adapted from
# https://github.com/joe-siyuan-qiao/Batch-Channel-Normalization
class BCNorm(nn.Module):
    def __init__(
        self, num_features: int, eps: float = 1e-7, estimate: bool = False
    ) -> None:
        """Batch channel normalization.

        https://arxiv.org/abs/1911.09738

        Infers the num_groups from the num_features to avoid
        errors. By default: uses 16 channels per group.
        If channels <= 16, squashes to batch layer norm

        magic number 16 comes from the paper:
        https://arxiv.org/abs/1803.08494

        Parameters
        ----------
            num_features : int
                Number of input channels/features.
            num_groups : int, default=None
                Number of groups to group the channels.
            eps : float, default=1e-7
                Small constant for numerical stability.
            estimate : bool, default=False
                If True, Uses `EstBN` instead of BN.
        """
        super().__init__()

        # Infer number of groups
        self.num_groups, remainder = divmod(num_features, 16)
        if remainder:
            self.num_groups = num_features // remainder

        self.num_features = num_features
        self.eps = eps
        self.weight = Parameter(torch.ones(1, self.num_groups, 1))
        self.bias = Parameter(torch.zeros(1, self.num_groups, 1))
        self.bn = nn.BatchNorm2d(num_features=num_features)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Batch-channel norm forward pass."""
        out = self.bn(x)
        out = out.reshape(1, x.size(0) * self.num_groups, -1)
        out = torch.batch_norm(out, None, None, None, None, True, 0, self.eps, True)
        out = out.reshape(x.size(0), self.num_groups, -1)
        out = self.weight * out + self.bias
        out = out.reshape_as(x)

        return out



from typing import Dict

import torch
import torch.nn as nn

from .joint_loss import JointLoss


class MultiTaskLoss(nn.ModuleDict):
    def __init__(
        self,
        branch_losses: Dict[str, JointLoss],
        loss_weights: Dict[str, float] = None,
        **kwargs,
    ) -> None:
        """Multi-task loss wrapper.

        Combines losses from different branches to one loss function.

        Parameters
        ----------
            branch_losses : Dict[str, nn.Module]
                Dictionary of branch names mapped to a loss module.
                e.g. {"inst": JointLoss(MSE(), Dice()), "type": Dice()}.
            loss_weights : Dict[str, float], default=None
                Dictionary of branch names mapped to the weight used for
                that branch loss.

        Raises
        ------
            ValueError:
                If the input arguments have different lengths.
                If the input arguments have mismatching keys.
        """
        super().__init__()

        self.weights = {k: 1.0 for k in branch_losses.keys()}
        if loss_weights is not None:
            if len(loss_weights) != len(branch_losses):
                raise ValueError(
                    f"""
                    Got {len(loss_weights)} loss weights and {len(branch_losses)}
                    branches. Need to have the same length."""
                )
            if not all(k in branch_losses.keys() for k in loss_weights.keys()):
                raise ValueError(
                    """Mismatching keys in `loss_weights` and `branch_losses`"""
                )
            else:
                self.weights = loss_weights

        for branch, loss in branch_losses.items():
            self.add_module(f"{branch}_loss", loss)

    def forward(
        self, yhats: Dict[str, torch.Tensor], targets: Dict[str, torch.Tensor], **kwargs
    ) -> torch.Tensor:
        """Compute the joint loss of the multi-task network.

        Parameters
        ----------
            yhats : Dict[str, torch.Tensor]
                Dictionary of branch names mapped to the predicted masks.
            targets : Dict[str, torch.Tensor]
                Dictionary of branch names mapped to the GT masks.

        Returns
        -------
            torch.Tensor:
                Computed multi-task loss (Scalar).
        """
        weight_map = None
        if "edgeweight" in targets.keys():
            weight_map = targets["edgeweight"]

        multitask_loss = 0.0
        for branch, loss in self.items():
            branch = branch.split("_")[0]
            branch_loss = loss(
                yhat=yhats[branch],
                target=targets[branch],
                target_weight=weight_map,
            )
            multitask_loss += branch_loss * self.weights[branch]

        return multitask_loss

    def extra_repr(self) -> str:
        """Add info to print."""
        s = "branch_loss_weights={weights}"
        return s.format(**self.__dict__)



from typing import Union

import numpy as np
import torch


def ndarray_to_tensor(
    array: np.ndarray, in_dim_format: str, out_dim_format: str
) -> torch.Tensor:
    """Convert img (H, W)|(H, W, C)|(B, H, W, C) to a tensor.

    Parameters
    ----------
        array : np.ndarray
            Numpy matrix. Shape: (H, W)|(H, W, C)|(B, H, W, C)
        in_dim_format : str
            The order of the dimensions in the input array.
            One of: "HW", "HWC", "BHWC", "BCHW", "BHW"
        out_dim_format : str
            The order of the dimensions in the output tensor.
            One of: "HW", "HWC", "BHWC", "BCHW", "BHW"

    Returns
    -------
        torch.Tensor:
            Input converted to a batched tensor. Shape .

    Raises
    ------
        TypeError:
            If input array is not np.ndarray.
        ValueError:
            If input array has wrong number of dimensions
        ValueError:
            If `in_dim_format` param is illegal.
        ValueError:
            If `out_dim_format` param is illegal.

    """
    if not isinstance(array, np.ndarray):
        raise TypeError(f"Input type: {type(array)} is not np.ndarray")

    if not 1 < len(array.shape) <= 4:
        raise ValueError(
            f"ndarray.shape {array.shape}, rank needs to be between [2, 4]"
        )

    dim_types = ("HW", "HWC", "BHWC", "BCHW", "BHW")
    if in_dim_format not in dim_types:
        raise ValueError(
            f"Illegal `in_dim_format`. Got {in_dim_format}. Allowed: {dim_types}"
        )

    if out_dim_format not in dim_types:
        raise ValueError(
            f"Illegal `out_dim_format`. Got {out_dim_format}. Allowed: {dim_types}"
        )

    if not len(array.shape) == len(in_dim_format):
        raise ValueError(
            f"""
            Mismatching input dimensions.
            Input Shape: {array.shape}.
            while `in_dim_format` is set to: {in_dim_format}"""
        )

    if in_dim_format in ("HW", "BHW"):
        if out_dim_format in ("HWC", "BHWC", "BCHW"):
            array = array[..., None]

    if in_dim_format in ("HW", "HWC"):
        if out_dim_format in ("BHWC", "BCHW", "BHW"):
            array = array[None, ...]

    if (
        len(array.shape) == 4
        and in_dim_format in ("BHWC", "HWC", "HW")
        and out_dim_format == "BCHW"
    ):
        array = array.transpose(0, 3, 1, 2)

    if len(array.shape) == 4 and in_dim_format == "BCHW" and out_dim_format == "BHWC":
        array = array.transpose(0, 2, 3, 1)

    return torch.from_numpy(array)


def tensor_to_ndarray(
    tensor: torch.Tensor, in_dim_format: str, out_dim_format: str
) -> np.ndarray:
    """Convert a tensor into a numpy ndarray.

    Parameters
    ----------
        tensor : torch.Tensor
            The input tensor. Shape: (B, H, W)|(B, C, H, W)
        in_dim_format : str
            The order of the dimensions in the input array.
            One of: "BCHW", "BHW"
        out_dim_format : str
            The order of the dimensions in the output tensor.
            One of: "HW", "HWC", "BHWC", "BHW"

    Returns
    -------
        np.ndarray:
            An ndarray. Shape(B, H, W, C)|(B, H, W)|(H, W, C)|(H, W)

    Raises
    ------
        TypeError:
            If input array is not torch.Tensor.
        ValueError:
            If input array has wrong number of dimensions
        ValueError:
            If `in_dim_format` param is illegal.
        ValueError:
            If `out_dim_format` param is illegal.
    """
    if not isinstance(tensor, torch.Tensor):
        raise TypeError(f"Input type: {type(tensor)} is not torch.Tensor")

    if not 3 <= tensor.dim() <= 4:
        raise ValueError(
            f"""
        The input tensor needs to have shape (B, H, W) or (B, C, H, W)."
        Got: {tensor.shape}""",
        )

    in_dim_types = ("BCHW", "BHW")
    if in_dim_format not in in_dim_types:
        raise ValueError(
            f"Illegal `in_dim_format`. Got {in_dim_format}. Allowed: {in_dim_types}"
        )

    out_dim_types = ("BCHW", "BHWC", "BHW", "HWC", "HW")
    if out_dim_format not in out_dim_types:
        raise ValueError(
            f"Illegal `out_dim_format`. Got {out_dim_format}. Allowed: {out_dim_types}"
        )

    # detach and bring to cpu
    array = tensor.detach()
    if tensor.is_cuda:
        array = array.cpu()

    array = array.numpy()
    if array.ndim == 4 and out_dim_format != "BCHW":
        array = array.transpose(0, 2, 3, 1)  # (B, H, W, C)

    if out_dim_format == "HW" and array.ndim == 4:
        array = array.squeeze()

    if out_dim_format == "HWC" and array.ndim == 4:
        try:
            array = array.squeeze(axis=0)
        except Exception:
            pass

    if out_dim_format == "BHW" and array.ndim == 4:
        try:
            array = array.squeeze(axis=-1)
        except Exception:
            pass

    if in_dim_format == "BHW":
        if out_dim_format == "BHWC":
            array = array[..., None]
        elif out_dim_format == "HW":
            array = array.squeeze()

    return array


def to_device(tensor: Union[torch.Tensor, np.ndarray]) -> torch.Tensor:
    """Push torch.Tensor or np.ndarray to GPU if it is available.

    Parameters
    ----------
        tensor : torch.Tensor or np.ndarray:
            Multi dim array to be pushed to gpu.

    Returns
    -------
        torch.Tensor:
            A tensor. Same shape as input.
    """
    if isinstance(tensor, np.ndarray):
        tensor = torch.from_numpy(tensor)

    if torch.cuda.is_available():
        tensor = tensor.cuda()

    return tensor


def tensor_one_hot(type_map: torch.Tensor, n_classes: int) -> torch.Tensor:
    """Convert a segmentation mask into one-hot-format.

    I.e. Takes in a segmentation mask of shape (B, H, W) and reshapes it
    into a tensor of shape (B, C, H, W).

    Parameters
    ----------
        type_map : torch.Tensor
            Multi-label Segmentation mask. Shape (B, H, W).
        n_classes : int
            Number of classes. (Zero-class included.)

    Returns
    -------
        torch.Tensor:
            A one hot tensor. Shape: (B, C, H, W). Dtype: torch.FloatTensor.

    Raises
    ------
        TypeError: If input is not torch.int64.
    """
    if not type_map.dtype == torch.int64:
        raise TypeError(
            f"""
            Input `type_map` should have dtype: torch.int64. Got: {type_map.dtype}."""
        )

    one_hot = torch.zeros(
        type_map.shape[0],
        n_classes,
        *type_map.shape[1:],
        device=type_map.device,
        dtype=type_map.dtype,
    )

    return one_hot.scatter_(dim=1, index=type_map.unsqueeze(1), value=1.0) + 1e-7



import numpy as np
import pytest

from cellseg_models_pytorch.metrics import (
    aggregated_jaccard_index,
    average_precision,
    dice2,
    iou_score,
    pairwise_object_stats,
    pairwise_pixel_stats,
    panoptic_quality,
)
from cellseg_models_pytorch.utils import remap_label


@pytest.fixture(scope="package")
def true() -> np.ndarray:
    true = np.array(
        [
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 1, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0],
            [0, 1, 1, 1, 1, 0, 0, 2, 2, 2, 2, 0, 0, 0],
            [1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 0, 0],
            [1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 0, 0],
            [0, 1, 1, 1, 1, 0, 0, 2, 2, 2, 2, 0, 0, 0],
            [0, 0, 1, 1, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 3, 3, 0, 0, 0, 0, 4, 4, 0],
            [0, 0, 0, 0, 3, 3, 3, 3, 0, 0, 4, 4, 4, 4],
            [0, 0, 0, 3, 3, 3, 3, 3, 3, 0, 4, 4, 4, 4],
            [0, 0, 0, 3, 3, 3, 3, 3, 3, 3, 0, 4, 4, 0],
            [0, 0, 0, 3, 3, 3, 3, 3, 3, 3, 0, 0, 0, 0],
        ],
        dtype=int,
    )

    return true


@pytest.fixture(scope="package")
def pred() -> np.ndarray:
    pred = np.array(
        [
            [0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 4, 4, 0, 0, 0, 0, 3, 3, 0, 0, 0, 0],
            [0, 4, 4, 4, 4, 0, 0, 3, 3, 3, 3, 0, 0, 0],
            [4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 0, 0],
            [4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 0, 0],
            [0, 4, 4, 4, 4, 0, 0, 3, 3, 3, 3, 0, 0, 0],
            [0, 0, 4, 4, 0, 0, 0, 0, 3, 3, 0, 0, 0, 0],
            [0, 0, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 5, 6, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 5, 5, 6, 6, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 5, 5, 5, 6, 6, 6, 0, 0, 0, 0, 0],
            [0, 0, 5, 5, 5, 5, 6, 6, 6, 6, 0, 0, 0, 0],
            [0, 5, 5, 5, 5, 5, 6, 6, 6, 6, 0, 0, 0, 0],
        ],
        dtype=int,
    )
    return pred


@pytest.mark.parametrize("metric_func", [None, iou_score])
def test_get_pixel_stats(true, pred, metric_func):
    expected_shape = (len(np.unique(true)[1:]), len(np.unique(pred)[1:]))
    stats = pairwise_pixel_stats(
        remap_label(true), remap_label(pred), metric_func=metric_func
    )[0]

    assert stats.shape == expected_shape


@pytest.mark.parametrize("sum_reduce", [True, False])
def test_pairwise_object_stats(true, pred, sum_reduce):
    expected_tp = 3
    expected_fp = 1
    expected_fn = 2

    iou = pairwise_pixel_stats(
        remap_label(true), remap_label(pred), metric_func=iou_score
    )
    iou = iou[0]
    matches = iou > 0.5

    tp_objects, fp_objects, fn_objects = pairwise_object_stats(matches)

    if sum_reduce:
        tp_objects = tp_objects.sum()
        fp_objects = fp_objects.sum()
        fn_objects = fn_objects.sum()

    assert tp_objects == expected_tp
    assert fp_objects == expected_fp
    assert fn_objects == expected_fn


@pytest.mark.parametrize("true_zeros", [None, np.zeros((14, 14))])
@pytest.mark.parametrize("pred_zeros", [None, np.zeros((14, 14))])
def test_panoptic_quality(true, pred, true_zeros, pred_zeros):

    if true_zeros is not None:
        true = true_zeros

    if pred_zeros is not None:
        pred = pred_zeros

    metrics = panoptic_quality(remap_label(true), remap_label(pred))

    expected = 0.0
    if pred_zeros is None and true_zeros is None:
        expected = 0.51

    np.testing.assert_allclose(
        metrics["pq"], expected, rtol=1e-2, err_msg=f"{metrics['pq']} != {expected}"
    )


@pytest.mark.parametrize("true_zeros", [None, np.zeros((14, 14))])
@pytest.mark.parametrize("pred_zeros", [None, np.zeros((14, 14))])
def test_average_precision(true, pred, true_zeros, pred_zeros):

    if true_zeros is not None:
        true = true_zeros

    if pred_zeros is not None:
        pred = pred_zeros

    ap = average_precision(remap_label(true), remap_label(pred))

    expected = 0.0
    if pred_zeros is None and true_zeros is None:
        expected = 0.6

    np.testing.assert_allclose(ap, expected, rtol=1e-2, err_msg=f"{ap} != {expected}")


@pytest.mark.parametrize("true_zeros", [None, np.zeros((14, 14))])
@pytest.mark.parametrize("pred_zeros", [None, np.zeros((14, 14))])
def test_dice2(true, pred, true_zeros, pred_zeros):

    if true_zeros is not None:
        true = true_zeros

    if pred_zeros is not None:
        pred = pred_zeros

    dice = dice2(remap_label(true), remap_label(pred))

    expected = 0.0
    if pred_zeros is None and true_zeros is None:
        expected = 0.64

    np.testing.assert_allclose(
        dice, expected, rtol=1e-2, err_msg=f"{dice} != {expected}"
    )


@pytest.mark.parametrize("true_zeros", [None, np.zeros((14, 14))])
@pytest.mark.parametrize("pred_zeros", [None, np.zeros((14, 14))])
def test_aji(true, pred, true_zeros, pred_zeros):

    if true_zeros is not None:
        true = true_zeros

    if pred_zeros is not None:
        pred = pred_zeros

    aji = aggregated_jaccard_index(remap_label(true), remap_label(pred))

    expected = 0.0
    if pred_zeros is None and true_zeros is None:
        expected = 0.53

    np.testing.assert_allclose(aji, expected, rtol=1e-2, err_msg=f"{aji} != {expected}")



import numpy as np
import pytest

from cellseg_models_pytorch.utils import (
    argmax,
    morph_chan_vese_thresh,
    naive_thresh,
    naive_thresh_prob,
    niblack_thresh,
    sauvola_thresh,
)


@pytest.fixture
def prob_map() -> np.ndarray:
    return np.random.rand(256, 256)


@pytest.mark.parametrize(
    "method",
    [
        naive_thresh,
        naive_thresh_prob,
        sauvola_thresh,
        morph_chan_vese_thresh,
        niblack_thresh,
    ],
)
def test_thresh(prob_map, method):
    binary = method(prob_map)

    if len(np.unique(binary)) == 2:
        assert np.amax(binary) == 1
        assert np.amin(binary) == 0

    assert binary.dtype == "uint8"


def test_argmax(prob_map):
    observed = argmax(prob_map)
    assert observed.shape == (256, 256)
    assert observed.dtype == "uint32"
